import 'dart:async';

import 'package:antenna/antenna.dart';

typedef Reducer<T> = T Function({T state, dynamic event});

abstract class Store<T> {
  T get state;

  StreamSubscription<T> listen(void Function(T state) stateListener);
}

class _Store<T> implements Store<T> {
  final Reducer<T> reducer;

  final subject = StreamController<T>.broadcast();

  late T _state;

  @override
  T get state => _state;

  set state(T value) {
    _state = value;

    subject.sink.add(value);
  }

  @override
  StreamSubscription<T> listen(void Function(T state) stateListener) {
    return subject.stream.listen(stateListener);
  }

  _Store(this.reducer) {
    state = reducer();
  }

  void dispatch(dynamic event) {
    final result = reducer(state: state, event: event);

    if (result == state) {
      return;
    }

    state = result;
  }
}

Store<T> createStore<T>(Reducer<T> reducer) => _Store<T>(reducer);

StreamSubscription connect<T>(Store<T> store) {
  final instance = store as _Store<T>;

  final subscription = listen(instance.dispatch);

  return subscription;
}
